
package eu.dnetlib.dhp.patchrefereed;

import static eu.dnetlib.dhp.common.SparkSessionSupport.runWithSparkSession;

import java.io.Serializable;
import java.util.List;
import java.util.Optional;

import org.apache.commons.io.IOUtils;
import org.apache.spark.SparkConf;
import org.apache.spark.api.java.function.MapFunction;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Encoders;
import org.apache.spark.sql.SaveMode;
import org.apache.spark.sql.SparkSession;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.ObjectMapper;

import eu.dnetlib.dhp.actionmanager.project.SparkAtomicActionJob;
import eu.dnetlib.dhp.application.ArgumentApplicationParser;
import eu.dnetlib.dhp.common.HdfsSupport;
import eu.dnetlib.dhp.schema.oaf.Instance;
import eu.dnetlib.dhp.schema.oaf.Qualifier;
import eu.dnetlib.dhp.schema.oaf.Result;
import scala.Tuple2;

public class SparkPatchRefereed implements Serializable {

	private static final Logger log = LoggerFactory.getLogger(SparkPatchRefereed.class);
	private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

	public static void main(String[] args) throws Exception {

		String jsonConfiguration = IOUtils
			.toString(
				SparkAtomicActionJob.class
					.getResourceAsStream(
						"/eu/dnetlib/dhp/patchrefereed/patch_parameters.json"));

		final ArgumentApplicationParser parser = new ArgumentApplicationParser(jsonConfiguration);

		parser.parseArgument(args);

		Boolean isSparkSessionManaged = Optional
			.ofNullable(parser.get("isSparkSessionManaged"))
			.map(Boolean::valueOf)
			.orElse(Boolean.TRUE);

		log.info("isSparkSessionManaged: {}", isSparkSessionManaged);

		final String outputPath = parser.get("outputPath");
		log.info("outputPath {}: ", outputPath);

		final String preparedInfoPath = parser.get("preparedInfoPath");
		log.info("preparedInfoPath {}: ", preparedInfoPath);

		final String inputPath = parser.get("inputPath");
		log.info("inputPath {}: ", inputPath);

		final String resultClassName = parser.get("resultTableName");
		log.info("resultTableName: {}", resultClassName);

		Class<? extends Result> resultClazz = (Class<? extends Result>) Class.forName(resultClassName);

		SparkConf conf = new SparkConf();

		runWithSparkSession(
			conf,
			isSparkSessionManaged,
			spark -> {
				removeOutputDir(spark, outputPath);
				patchRefereed(
					spark,
					inputPath,
					preparedInfoPath,
					outputPath,
					resultClazz);
			});
	}

	private static void removeOutputDir(SparkSession spark, String path) {
		HdfsSupport.remove(path, spark.sparkContext().hadoopConfiguration());
	}

	private static <R extends Result> void patchRefereed(SparkSession spark,
		String inputPath,
		String prearedInfoPath,
		String outputPath,
		Class<R> resultClazz) {

		Dataset<R> result = readPath(spark, inputPath, resultClazz);
		Dataset<ResultInstance> ri = readPath(spark, prearedInfoPath, ResultInstance.class);

		result
			.joinWith(ri, result.col("id").equalTo(ri.col("resultId")), "left")
			.map((MapFunction<Tuple2<R, ResultInstance>, R>) value -> {
				return updateRefereed(value);
			}, Encoders.bean(resultClazz))
			.write()
			.mode(SaveMode.Overwrite)
			.option("compression", "gzip")
			.json(outputPath);

	}

	private static <R extends Result> R updateRefereed(Tuple2<R, ResultInstance> value) {
		R r = value._1();
		Optional<ResultInstance> oInstanceList = Optional.ofNullable(value._2());
		Optional<List<Instance>> oresultInstance = Optional.ofNullable(r.getInstance());
		if (!oresultInstance.isPresent()) {
			return r;
		}
		if (oInstanceList.isPresent()) {
			List<Instance> instanceList = oInstanceList.get().getInstanceList();
			oresultInstance.get().forEach(i -> checkEquivalence(instanceList, i));
		} else {
			oresultInstance.get().forEach(i -> i.setRefereed(Constants.DEFAULT_REFEREED));
		}
		return r;
	}

	private static void checkEquivalence(List<Instance> betaResultInstance, Instance i) {

		for (Instance ri : betaResultInstance) {
			if (update(ri, i)) {
				return;
			}
		}
		i.setRefereed(Constants.DEFAULT_REFEREED);

	}

	// verify collectedfrom hostedby openaccess instancetype url
	private static boolean update(Instance ri, Instance i) {
		if (// ri.getAccessright().equals(i.getAccessright()) &&
		ri.getCollectedfrom().equals(i.getCollectedfrom()) &&
		// ri.getHostedby().equals(i.getHostedby()) &&
		// ri.getInstancetype().equals(i.getInstancetype()) &&
			equals(ri.getUrl(), i.getUrl())) {
			// Optional<Qualifier> oRefereed = Optional.ofNullable(i.getRefereed());
			// if (!oRefereed.isPresent()) {
			// if (!(ri.getRefereed().getClassid().equals("") ||
			// ri.getRefereed().getClassid().equals(Constants.DEFAULT_REFEREED.getClassid()))) {
			i.setRefereed(ri.getRefereed());
			return true;
			// }
		}
		return false;
	}

	// TODO verify the reason of the NullPointerException fired at line 167
	public static boolean equals(List<String> url, List<String> url1) {
		try {
			if (url == null && url1 == null) {
				return true;
			}

			if ((url1 == null && url != null) || (url != null && url1 == null)) {
				return false;
			}

			if (url.size() != url1.size()) {
				return false;
			}
			for (String u : url) {
				if (!url1.contains(u)) {
					return false;
				}
			}
			return true;
		} catch (Exception e) {
			return false;
		}

	}

	public static <R> Dataset<R> readPath(
		SparkSession spark, String inputPath, Class<R> clazz) {
		return spark
			.read()
			.textFile(inputPath)
			.map((MapFunction<String, R>) value -> OBJECT_MAPPER.readValue(value, clazz), Encoders.bean(clazz));
	}

}
