
package eu.dnetlib.dhp.patchrefereed;

import static eu.dnetlib.dhp.common.SparkSessionSupport.runWithSparkSession;

import java.io.Serializable;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.commons.io.IOUtils;
import org.apache.spark.SparkConf;
import org.apache.spark.api.java.function.MapFunction;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Encoders;
import org.apache.spark.sql.SaveMode;
import org.apache.spark.sql.SparkSession;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.ObjectMapper;

import eu.dnetlib.dhp.application.ArgumentApplicationParser;
import eu.dnetlib.dhp.common.HdfsSupport;
import eu.dnetlib.dhp.schema.oaf.*;

public class SparkPrepareResultInstanceList implements Serializable {

	private static final Logger log = LoggerFactory.getLogger(SparkPrepareResultInstanceList.class);
	private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

	public static void main(String[] args) throws Exception {

		String jsonConfiguration = IOUtils
			.toString(
				SparkPrepareResultInstanceList.class
					.getResourceAsStream(
						"/eu/dnetlib/dhp/patchrefereed/prepare_parameters.json"));

		final ArgumentApplicationParser parser = new ArgumentApplicationParser(jsonConfiguration);

		parser.parseArgument(args);

		Boolean isSparkSessionManaged = Optional
			.ofNullable(parser.get("isSparkSessionManaged"))
			.map(Boolean::valueOf)
			.orElse(Boolean.TRUE);

		log.info("isSparkSessionManaged: {}", isSparkSessionManaged);

		final String outputPath = parser.get("outputPath");
		log.info("outputPath {}: ", outputPath);

		final String inputPath = parser.get("inputPath");
		log.info("inputPath {}: ", inputPath);

		final String resultClassName = parser.get("resultTableName");
		log.info("resultTableName: {}", resultClassName);

		Class<? extends Result> resultClazz = (Class<? extends Result>) Class.forName(resultClassName);

		SparkConf conf = new SparkConf();

		runWithSparkSession(
			conf,
			isSparkSessionManaged,
			spark -> {
				removeOutputDir(spark, outputPath);
				getAtomicActions(
					spark,
					inputPath,
					outputPath,
					resultClazz);
			});
	}

	private static void removeOutputDir(SparkSession spark, String path) {
		HdfsSupport.remove(path, spark.sparkContext().hadoopConfiguration());
	}

	private static <R extends Result> void getAtomicActions(SparkSession spark,
		String inputPath,
		String outputPath,
		Class<R> resultClazz) {

		Dataset<R> result = readPath(spark, inputPath, resultClazz);

		// log.info("number of results: {}", result.count());
		result.map(r -> {
			ResultInstance ri = null;
			List<Instance> instanceList = Optional
				.ofNullable(r.getInstance())
				.map(
					instances -> instances
						.stream()
						.map(instance -> {
							Optional<Qualifier> oRefereed = Optional.ofNullable(instance.getRefereed());
							if (oRefereed.isPresent()) {
								if (!(oRefereed.get().getClassid().equals(Constants.DEFAULT_REFEREED.getClassid()) ||
									oRefereed.get().getClassid().equals("UNKNOWN") ||
									oRefereed.get().getClassid().equals(""))) {
									return instance;
								}
							}
							return null;
						})
						.filter(Objects::nonNull)
						.collect(Collectors.toList()))
				.orElse(null);

			if ((instanceList != null) && instanceList.size() > 0) {
				ri = new ResultInstance();
				ri.setResultId(r.getId());
				ri.setInstanceList(instanceList);
			}

			return ri;

		}, Encoders.bean(ResultInstance.class))
			.filter(Objects::nonNull)
			.write()
			.mode(SaveMode.Overwrite)
			.option("compression", "gzip")
			.json(outputPath);
	}

	public static <R> Dataset<R> readPath(
		SparkSession spark, String inputPath, Class<R> clazz) {
		return spark
			.read()
			.textFile(inputPath)
			.map((MapFunction<String, R>) value -> OBJECT_MAPPER.readValue(value, clazz), Encoders.bean(clazz));
	}

}
